﻿namespace Jsl.FxCop
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Microsoft.FxCop.Sdk;

    /// <summary>
    /// A helper class for working with Microsoft FxCop reflection.
    /// </summary>
    public static class NodeHelper
    {
        #region Public Methods
        /// <summary>
        /// Gets the fields for the specified type.
        /// </summary>
        /// <param name="targetType">Type of the target.</param>
        /// <returns>The fields.</returns>
        public static IEnumerable<Field> GetFieldNodes(this Type targetType)
        {
            return targetType.GetTypeNode().Members.OfType<Field>();
        }

        /// <summary>
        /// Gets the method.
        /// </summary>
        /// <param name="code">The code to get the <see cref="Method"/> for.</param>
        /// <returns>The method found.</returns>
        public static Method GetMethod(Action code)
        {
            if (code == null)
            {
                throw new ArgumentNullException("code");
            }

            return
                (from method in code.Method.DeclaringType.GetMethodNodes()
                 where method.Name.Name == code.Method.Name
                 select method).Single();
        }

        /// <summary>
        /// Gets the method.
        /// </summary>
        /// <typeparam name="TValue">The type of the value.</typeparam>
        /// <param name="code">The code to get the <see cref="Method"/> for.</param>
        /// <returns>The method found.</returns>
        public static Method GetMethod<TValue>(Func<TValue> code)
        {
            if (code == null)
            {
                throw new ArgumentNullException("code");
            }

            return
                (from method in code.Method.DeclaringType.GetMethodNodes()
                 where method.Name.Name == code.Method.Name
                 select method).Single();
        }

        /// <summary>
        /// Gets the single method.
        /// </summary>
        /// <param name="targetType">Type of the target.</param>
        /// <param name="predicate">The predicate to determine the method to return.</param>
        /// <returns>The method.</returns>
        public static Method GetMethod(this Type targetType, Func<Method, bool> predicate)
        {
            return
                targetType.GetMethodNodes()
                    .Where(predicate)
                    .Single();
        }

        /// <summary>
        /// Gets the method nodes.
        /// </summary>
        /// <param name="targetType">Type of the target.</param>
        /// <returns>The methods found.</returns>
        public static IEnumerable<Method> GetMethodNodes(this Type targetType)
        {
            return targetType.GetTypeNode().Members.OfType<Method>();
        }

        /// <summary>
        /// Gets the property nodes for the specified type.
        /// </summary>
        /// <param name="targetType">Type of the target.</param>
        /// <returns>The property nodes.</returns>
        public static IEnumerable<PropertyNode> GetPropertyNodes(this Type targetType)
        {
            return targetType.GetTypeNode().Members.OfType<PropertyNode>();
        }
        #endregion Public Methods
    }
}
